#  Copyright (c) 2017 Red Hat, Inc.
#
#  This file is part of ARA: Ansible Run Analysis.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from flask import Blueprint
from flask import abort
from flask import current_app
from flask import jsonify
from flask import redirect
from flask import render_template
from flask import url_for

from ara import utils
from ara.db import models

reports = Blueprint('reports', __name__)


@reports.route('/')
@reports.route('/reports')
@reports.route('/reports/list/<int:page>')
def report_list(page=1):
    playbooks = models.Playbook.query.order_by(models.Playbook.started.desc())

    if not utils.fast_count(playbooks):
        return redirect(url_for('about.main'))

    playbook_per_page = current_app.config['ARA_PLAYBOOK_PER_PAGE']
    # Paginate unless playbook_per_page is set to 0
    if playbook_per_page >= 1:
        playbooks = playbooks.paginate(page, playbook_per_page, False)
    else:
        playbooks = playbooks.paginate(page, None, False)

    stats = utils.get_summary_stats(playbooks.items, 'playbook_id')

    result_per_page = current_app.config['ARA_RESULT_PER_PAGE']

    return render_template('report_list.html',
                           active='reports',
                           result_per_page=result_per_page,
                           playbooks=playbooks,
                           stats=stats)


@reports.route('/reports/<playbook_id>')
def report(playbook_id):
    playbook = models.Playbook.query.get(playbook_id)
    if playbook is None:
        abort(404)

    stats = utils.get_summary_stats([playbook], 'playbook_id')

    result_per_page = current_app.config['ARA_RESULT_PER_PAGE']

    return render_template('report_single.html',
                           active='reports',
                           result_per_page=result_per_page,
                           playbook=playbook,
                           stats=stats)

# Note (dmsimard)
# We defer the loading of the different data tables and render them
# asynchronously for performance purposes.
# Because of this, we need to prepare some JSON data so data tables can
# retrieve it through AJAX:
#   https://datatables.net/examples/ajax/defer_render.html
# The following routes are there to support this mechanism.
# The routes have a text extension for proper mimetype detection.


@reports.route('/reports/ajax/parameters/<playbook>')
def ajax_parameters(playbook):
    playbook = models.Playbook.query.get(playbook)
    if playbook is None:
        abort(404)

    results = dict()
    results['data'] = list()

    results['data'].append(['playbook_path', playbook.path])
    results['data'].append(['ansible_version', playbook.ansible_version])

    if playbook.parameters:
        for parameter in playbook.parameters:
            results['data'].append([parameter, playbook.parameters[parameter]])

    return jsonify(results)


@reports.route('/reports/ajax/plays/<playbook>')
def ajax_plays(playbook):
    plays = (models.Play.query
             .filter(models.Play.playbook_id.in_([playbook])))
    if not utils.fast_count(plays):
        abort(404)

    jinja = current_app.jinja_env
    date = jinja.from_string('{{ date | datefmt }}')
    time = jinja.from_string('{{ time | timefmt }}')

    results = dict()
    results['data'] = list()

    for play in plays:
        name = u"<span class='pull-left'>{0}</span>".format(play.name)
        start = date.render(date=play.started)
        end = date.render(date=play.ended)
        duration = time.render(time=play.duration)
        results['data'].append([name, start, end, duration])

    return jsonify(results)


@reports.route('/reports/ajax/records/<playbook>')
def ajax_records(playbook):
    records = (models.Record.query
               .filter(models.Record.playbook_id.in_([playbook])))
    if not utils.fast_count(records):
        abort(404)

    jinja = current_app.jinja_env
    record_key = jinja.get_template('ajax/record_key.html')
    record_value = jinja.get_template('ajax/record_value.html')

    results = dict()
    results['data'] = list()

    for record in records:
        key = record_key.render(record=record)
        value = record_value.render(record=record)

        results['data'].append([key, value])

    return jsonify(results)


@reports.route('/reports/ajax/results/<playbook>')
def ajax_results(playbook):
    results = (models.Result.query
               .join(models.Task)
               .filter(models.Task.playbook_id.in_([playbook])))
    if not utils.fast_count(results):
        abort(404)

    jinja = current_app.jinja_env
    time = jinja.from_string('{{ time | timefmt }}')
    action_link = jinja.get_template('ajax/action.html')
    name_cell = jinja.get_template('ajax/task_name.html')
    task_status_link = jinja.get_template('ajax/task_status.html')

    ajax = dict()
    ajax['data'] = list()

    for result in results:
        name = name_cell.render(result=result)
        host = result.host.name
        action = action_link.render(result=result)
        elapsed = time.render(time=result.task.offset_from_playbook)
        duration = time.render(time=result.duration)
        status = task_status_link.render(result=result)

        ajax['data'].append([name, host, action, elapsed, duration, status])
    return jsonify(ajax)


@reports.route('/reports/ajax/hosts/<playbook>')
def ajax_hosts(playbook):
    hosts = (models.Host.query
             .filter(models.Host.playbook_id.in_([playbook])))
    if not utils.fast_count(hosts):
        abort(404)

    jinja = current_app.jinja_env
    host_link = jinja.get_template('ajax/host.html')

    results = dict()
    results['data'] = list()

    for host in hosts:
        host_render = host_link.render(host=host)
        ok = host.ok if host.ok >= 1 else 0
        changed = host.changed if host.changed >= 1 else 0
        failed = host.failed if host.failed >= 1 else 0
        skipped = host.skipped if host.skipped >= 1 else 0
        unreachable = host.unreachable if host.unreachable >= 1 else 0

        data = [host_render, ok, changed, failed, skipped, unreachable]
        results['data'].append(data)

    return jsonify(results)
